package com.craftaro.core;

import com.craftaro.core.configuration.Config;
import com.craftaro.core.database.DataManager;
import com.craftaro.core.database.DataMigration;
import com.craftaro.core.database.DatabaseType;
import com.craftaro.core.locale.Locale;
import com.craftaro.core.utils.Metrics;
import com.craftaro.core.utils.SongodaAuth;
import de.tr7zw.changeme.nbtapi.utils.MinecraftVersion;
import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.command.CommandSender;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.plugin.java.JavaPlugin;

import java.io.File;
import java.sql.Connection;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;

/**
 * REMINDER: When converting plugins to use this, REMOVE METRICS <br>
 * Must not have two instances of Metrics enabled!
 */
public abstract class CraftaroPlugin extends JavaPlugin {
    protected Locale locale;
    protected Config config = new Config(this);
    protected Config databaseConfig;
    protected DataManager dataManager;
    protected long dataLoadDelay = 20L;

    private boolean emergencyStop = false;

    static {
        /* NBT-API */
        MinecraftVersion.getLogger().setLevel(Level.WARNING);
        MinecraftVersion.disableUpdateCheck();
        // Disable tips and logo for Jooq
        System.setProperty("org.jooq.no-tips", "true");
        System.setProperty("org.jooq.no-logo", "true");
    }

    public abstract void onPluginLoad();

    public abstract void onPluginEnable();

    public abstract void onPluginDisable();

    public abstract void onDataLoad();

    /**
     * Called after reloadConfig() is called
     */
    public abstract void onConfigReload();

    /**
     * Any other plugin configuration files used by the plugin.
     *
     * @return a list of Configs that are used in addition to the main config.
     */
    public abstract List<Config> getExtraConfig();

    @Override
    public FileConfiguration getConfig() {
        return config.getFileConfig();
    }

    public Config getCoreConfig() {
        return config;
    }

    @Override
    public void reloadConfig() {
        config.load();
        onConfigReload();
    }

    @Override
    public void saveConfig() {
        config.save();
    }

    @Override
    public final void onLoad() {
        try {
            onPluginLoad();
        } catch (Throwable th) {
            criticalErrorOnPluginStartup(th);
        }
    }

    @Override
    public final void onEnable() {
        if (this.emergencyStop) {
            setEnabled(false);

            return;
        }

        CommandSender console = Bukkit.getConsoleSender();

        // Check plugin access, don't load plugin if user don't have access
        if (!SongodaAuth.isAuthorized(true)) {
            String pluginName = getDescription().getName();

            new Thread(() -> {
                String externalIP = SongodaAuth.getIP();
                String serverUuid = SongodaAuth.getUUID().toString();

                String message = "\n" +
                        ChatColor.RED + "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" +
                        ChatColor.RED + "You do not have access to the " + pluginName + " plugin.\n" +
                        ChatColor.YELLOW + "Please purchase a license at https://sngda.to/marketplace\n" +
                        ChatColor.YELLOW + "or set up your license at https://sngda.to/licenses\n" +
                        ChatColor.YELLOW + "License setup steps:\n" +
                        ChatColor.YELLOW + "Visit the link mentioned above and click the 'Create License' button.\n" +
                        ChatColor.YELLOW + "Copy the following IP address and UUID and click create.\n" +
                        ChatColor.YELLOW + "UUID: " + serverUuid + "\n" +
                        ChatColor.YELLOW + "IP: " + externalIP + "\n" +
                        ChatColor.RED + "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
                console.sendMessage(message);
            }).start();

            emergencyStop();
            return;
        }

        console.sendMessage(" "); // blank line to separate chatter
        console.sendMessage(ChatColor.GREEN + "=============================");
        console.sendMessage(String.format("%s%s %s by %sCraftaro <3!", ChatColor.GRAY,
                getDescription().getName(), getDescription().getVersion(), ChatColor.DARK_PURPLE));
        console.sendMessage(String.format("%sAction: %s%s%s...", ChatColor.GRAY,
                ChatColor.GREEN, "Enabling", ChatColor.GRAY));

        try {
            this.locale = Locale.loadDefaultLocale(this, "en_US");

            // plugin setup
            onPluginEnable();

            if (this.emergencyStop) {
                return;
            }

            // Load Data.
            Bukkit.getScheduler().runTaskLater(this, this::onDataLoad, this.dataLoadDelay);

            if (this.emergencyStop) {
                console.sendMessage(ChatColor.RED + "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
                console.sendMessage(" ");
                return;
            }

            // Start Metrics
            Metrics.start(this);
        } catch (Throwable th) {
            criticalErrorOnPluginStartup(th);

            console.sendMessage(ChatColor.RED + "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
            console.sendMessage(" ");

            return;
        }

        console.sendMessage(ChatColor.GREEN + "=============================");
        console.sendMessage(" "); // blank line to separate chatter
    }

    @Override
    public final void onDisable() {
        if (this.emergencyStop) {
            return;
        }

        CommandSender console = Bukkit.getConsoleSender();

        console.sendMessage(" "); // blank line to separate chatter
        console.sendMessage(ChatColor.GREEN + "=============================");
        console.sendMessage(String.format("%s%s %s by %sCraftaro <3!", ChatColor.GRAY,
                getDescription().getName(), getDescription().getVersion(), ChatColor.DARK_PURPLE));
        console.sendMessage(String.format("%sAction: %s%s%s...", ChatColor.GRAY,
                ChatColor.RED, "Disabling", ChatColor.GRAY));

        onPluginDisable();
        try (Connection connection = this.dataManager.getDatabaseConnector().getConnection()) {
            System.err.println("Connection not closed! Closing...");
            connection.close();
            this.dataManager.getDatabaseConnector().closeConnection();
        } catch (Exception ignored) {
            ignored.printStackTrace();
        }

        console.sendMessage(ChatColor.GREEN + "=============================");
        console.sendMessage(" "); // blank line to separate chatter
    }

    public Locale getLocale() {
        return this.locale;
    }

    /**
     * Set the plugin's locale to a specific language
     *
     * @param localeName locale to use, eg "en_US"
     * @param reload     optionally reload the loaded locale if the locale didn't
     *                   change
     *
     * @return true if the locale exists and was loaded successfully
     */
    public boolean setLocale(String localeName, boolean reload) {
        if (this.locale != null && this.locale.getName().equals(localeName)) {
            return !reload || this.locale.reloadMessages();
        }

        Locale l = Locale.loadLocale(this, localeName);
        if (l != null) {
            this.locale = l;
            return true;
        }

        return false;
    }

    protected void emergencyStop() {
        this.emergencyStop = true;

        Bukkit.getPluginManager().disablePlugin(this);
    }

    /**
     * Logs one or multiple errors that occurred during plugin startup and calls {@link #emergencyStop()} afterwards
     *
     * @param th The error(s) that occurred
     */
    protected void criticalErrorOnPluginStartup(Throwable th) {
        Bukkit.getLogger().log(Level.SEVERE,
                String.format(
                        "Unexpected error while loading %s v%s (core v%s): Disabling plugin!",
                        getDescription().getName(),
                        getDescription().getVersion(),
                        CraftaroCore.getVersion()
                ), th);

        emergencyStop();
    }

    //New database stuff
    public Config getDatabaseConfig() {
        File databaseFile = new File(getDataFolder(), "database.yml");
        if (!databaseFile.exists()) {
            saveResource("database.yml", false);
        }
        if (this.databaseConfig == null) {
            this.databaseConfig = new Config(databaseFile);
            this.databaseConfig.load();
        }
        return this.databaseConfig;
    }

    /**
     * Get the DataManager for this plugin.
     * Note: Make sure to call initDatabase() in onPluginEnable() before using this.
     * @return DataManager for this plugin.
     */
    public DataManager getDataManager() {
        return dataManager;
    }

    /**
     * Initialize the DataManager for this plugin and convert from SQLite to H2 if needed.
     */
    protected void initDatabase() {
        initDatabase(Collections.emptyList());
    }

    /**
     * Initialize the DataManager for this plugin and convert from SQLite to H2 if needed.
     * @param migrations List of migrations to run.
     */
    protected void initDatabase(List<DataMigration> migrations) {
        boolean legacy = this.config.contains("MySQL");
        boolean isSQLite = !this.config.getBoolean("MySQL.Enabled", false);
        if (legacy && isSQLite) {
            this.config.set("MySQL", null);
            this.dataManager = new DataManager(this,migrations, DatabaseType.SQLITE);
        } else if (legacy) {
            //Copy creditental from old config to new config
            this.databaseConfig.set("MySQL.Hostname", this.config.getString("MySQL.Hostname", "localhost"));
            this.databaseConfig.set("MySQL.Port", this.config.getInt("MySQL.Port", 3306));
            this.databaseConfig.set("MySQL.Database", this.config.getString("MySQL.Database", "database"));
            this.databaseConfig.set("MySQL.Username", this.config.getString("MySQL.Username", "username"));
            this.databaseConfig.set("MySQL.Password", this.config.getString("MySQL.Password", "password"));
            this.databaseConfig.set("MySQL.Pool Size", this.config.getInt("MySQL.Pool Size", 5));
            this.databaseConfig.set("MySQL.Use SSL", this.config.getBoolean("MySQL.Use SSL", false));
            this.dataManager = new DataManager(this, migrations);
        } else {
            this.dataManager = new DataManager(this, migrations);
        }
        if (dataManager.getDatabaseConnector().isInitialized()) {
            //Check if the type is SQLite
            if (dataManager.getDatabaseConnector().getType() == DatabaseType.SQLITE) {
                //Let's convert it to H2
                DataManager newDataManager = DataMigration.convert(this, DatabaseType.H2);
                if (newDataManager != null && newDataManager.getDatabaseConnector().isInitialized()) {
                    //Set the new data manager
                    setDataManager(newDataManager);
                }
            }
        }
    }

    /**
     * Set the DataManager for this plugin.
     * Used for converting from one database to another.
     */
    public void setDataManager(DataManager dataManager) {
        if (dataManager == null) throw new IllegalArgumentException("DataManager cannot be null!");
        if (this.dataManager == dataManager) return;
        //Make sure to shut down the old data manager.
        if (this.dataManager != null) {
            dataManager.shutdown();
        }
        this.dataManager = dataManager;
    }
}
